#include <stdio.h>
#include <string.h>
#include "expat_module.h"
#include "parse_event_handler.h"
#include "util.h"

/** EXPAT XML_Char to PyUnicode helpers *********************************/

static PyObject *Unicode_FromXMLCharAndSize(const XML_Char *str, int size)
{
  /* sizeof(Py_UNICODE) could be 4, however UTF-16 is 2 bytes wide */
  return PyUnicode_DecodeUTF16((const char *)str, size<<1, NULL, NULL);
}

static PyObject *Unicode_FromXMLChar(const XML_Char *str)
{
  const XML_Char *p = str;

  while (*p) p++;
  /* sizeof(Py_UNICODE) could be 4, however UTF-16 is 2 bytes wide */
  return PyUnicode_DecodeUTF16((const char *)str, (p - str)<<1, NULL, NULL);
}

/** EXPAT callbacks *****************************************************/

extern int entityRef(XML_Parser parser, const XML_Char *context,
                     const XML_Char *base, const XML_Char *systemId,
                     const XML_Char *publicId);
static void comment(void *userData, const XML_Char *data);

static void startElement(void *userData, const XML_Char *name, 
                         const XML_Char **atts);
static void endElement(void *userData, const XML_Char *name);

static void startNsScope(void *userData, const XML_Char *prefix,
                         const XML_Char *uri);
static void endNsScope(void *userData, const XML_Char *prefix);

static void characterData(void *userData, const XML_Char *s, int len);

static void processingInstruction(void *userData, const XML_Char *target,
                                  const XML_Char *data);

static void startDoctypeDecl(void *userData, const XML_Char *name,
                             const XML_Char *sysid, const XML_Char *pubid,
                             int has_internal_subset);
static void endDoctypeDecl(void *userData);

int initializeStateTableForParseEvents(ParserState *state);


int setupParserState(ParserState **sptr, PyDocumentObject *doc, 
                     PyObject *inputSource)
{
  ParserState *s;
  Stack *ps, *ns, *xbs, *edes;
  List *nss, *vhrefs;
  int *new_preserve_state;

  if ((s = (ParserState *)calloc(1, sizeof(ParserState))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  if ((edes = (Stack *)malloc(sizeof(Stack))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  _stack_init(edes, free);
  s->elem_depth_event_stack = edes;
  if ((ps = (Stack *)malloc(sizeof(Stack))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  _stack_init(ps, free);
  if ((new_preserve_state = (int *)malloc(sizeof(int))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  *new_preserve_state = 1;
  _stack_push(ps, ((void *)new_preserve_state));
  s->preserve_state = ps;

  if ((ns = (Stack *)malloc(sizeof(Stack))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  /*We just borrow reference to these objects so there is no need to DECREF when done*/
  _stack_init(ns, NULL);
  s->node_stack = ns;

  /* new namespaces */
  if ((s->new_nss = PyDict_New()) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }

  /* in-scope namespaces */
  if ((nss = (List *)malloc(sizeof(List))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  list_init(nss, free_nsmapping);
  s->namespaces = nss;

  /* XML Base stack */
  if ((xbs = (Stack *)malloc(sizeof(Stack))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  /* FIXME: memory leak.  Fix by creating function that simply does Py_DECREF */
  _stack_init(xbs, NULL);
  _stack_push(xbs, ((void *)PyObject_GetAttrString(inputSource,"uri")));
  s->xml_base_stack = xbs;

  s->ownerDoc = doc;

  _stack_push(s->node_stack, doc);

  s->curr_text = NULL;

  s->xinclude_content_depth = 0;
  s->input_source = inputSource;

  if ((vhrefs = (List *)malloc(sizeof(List))) == NULL) {
    /* Exception: Out of memory */
    return 0;
  }
  list_init(vhrefs, NULL);
  s->visited_hrefs = vhrefs;

  /* set up state table */
  if (!initializeStateTable(s))
    return 0;
  if (newStateTableEntry(s, START_STATE) == NULL)
    return 0;
  if (newStateTableEntry(s, PARSE_STREAM_STATE) == NULL)
    return 0;
  if (!addTransition(s, START_STATE, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;
  if (!addTransition(s, PARSE_STREAM_STATE, PARSE_RESUME_EVENT,
                     PARSE_STREAM_STATE))
    return 0;
  if (!initializeStateTableForParseEvents(s))
    return 0;

  *sptr = s;
  return 1;
}


int cleanupParserState(ParserState *s)
{
  void *junk;

  destroyStateTable(s);

  _stack_pop(s->node_stack, &junk);

  _stack_pop(s->preserve_state, &junk);
  free(junk);

  _stack_destroy(s->preserve_state);
  _stack_destroy(s->node_stack);
  _stack_destroy(s->xml_base_stack);
  _stack_destroy(s->elem_depth_event_stack);
  list_destroy(s->namespaces);

  free(s->visited_hrefs);
  free(s->preserve_state);
  free(s->node_stack);
  free(s->xml_base_stack);
  free(s->namespaces);
  free(s->elem_depth_event_stack);

  Py_DECREF(s->new_nss);
  Py_XDECREF(s->curr_text);
  free(s);
  return 1;
}


int initParser(XML_Parser *parser, ParserState *state)
{
  PyObject *unicodeBase,*base;
  XML_Char EMPTY_WSTR[] = {'\0'};
  XML_SetUserData(*parser, state);

  /* enable prefix information in names (URI + sep + local + sep + prefix) */
  XML_SetReturnNSTriplet(*parser, 1);
  XML_SetStartNamespaceDeclHandler(*parser, startNsScope);

  XML_SetElementHandler(*parser, startElement, endElement);
  XML_SetCharacterDataHandler(*parser, characterData);
  XML_SetProcessingInstructionHandler(*parser, processingInstruction);
  XML_SetCommentHandler(*parser, comment);

  /* Capture doctype events to disable comments and processing instructions
     while reading the DTD,
  */
  XML_SetStartDoctypeDeclHandler(*parser, startDoctypeDecl);
  XML_SetEndDoctypeDeclHandler(*parser, endDoctypeDecl);

  XML_SetExternalEntityRefHandler(*parser, entityRef);

  base = PyObject_GetAttrString(state->input_source,"uri");
  if (base == NULL) {
    return 0;
  };

  unicodeBase = PyUnicode_FromObject(base);
  if (!unicodeBase) {
    return 0;
  }

  if (!XML_SetBase(*parser, base ? (XML_Char *)PyUnicode_AS_UNICODE(unicodeBase) : EMPTY_WSTR))
  {
    Py_DECREF(unicodeBase);
    return 0;
  }
  Py_DECREF(unicodeBase);
  Py_DECREF(base);

  state->in_dtd = 0;
  return 1;
}


PyObject *beginParse(PyObject *inputSource, int readDtd)
{
  char buf[BUFSIZ];
  XML_Parser parser = XML_ParserCreateNS(NULL, EXPAT_NSSEP);
  int done;
  size_t len;
  ParserState *state;
  unsigned long *docIx;
  PyDocumentObject *doc = NULL;
  PyObject *stream, *base,*baseUnicode;

  if (readDtd) XML_SetParamEntityParsing(parser, XML_PARAM_ENTITY_PARSING_ALWAYS);

  if ((docIx = (unsigned long *)malloc(sizeof(unsigned long))) == NULL) {
    /* Exception: Out of memory */
  }
  *docIx = 0;
  base = PyObject_GetAttrString(inputSource, "uri");
  if (!base) {
    return NULL;
  }

  baseUnicode = PyUnicode_FromObject(base);
  Py_DECREF(base);
  if (!baseUnicode) return NULL;

  if ((doc = Document_NEW(docIx, baseUnicode)) == NULL) {
    /* Exception: Out of memory */
    Py_DECREF(baseUnicode);
    return NULL;
  }
  Py_DECREF(baseUnicode);

  if (!setupParserState(&state, doc, inputSource)){
    /* an error occurred */
    Py_DECREF(doc);
    return NULL;
  }
  state->docIx = docIx;
  resetState(state, START_STATE);

  if (!initParser(&parser, state))
  {
    Py_DECREF(doc);
    return NULL;
  }

  /*Pre-get the stream*/
  stream = PyObject_GetAttrString(inputSource, "stream");
  if (stream == NULL) {
    Py_DECREF(doc);
    return NULL;
  }

  do {
    len = readFromObject(stream, buf, BUFSIZ);
    if (PyErr_Occurred()) {
      Py_DECREF(doc);
      Py_DECREF(stream);
      return NULL;
    }
    done = len < sizeof(buf);
    transit(state, PARSE_RESUME_EVENT);
    if (!XML_Parse(parser, buf, len, done)) {
      /* Exception: Well-formedness error */
      /*PyErr_Format(g_readerErrorObject, ...);*/
      if (!PyErr_Occurred()) {
        PyObject *uri = PyObject_GetAttrString(inputSource, "uri");
        PyObject *repr = PyObject_Repr(uri);
        Py_DECREF(uri);
        PyErr_Format(PyExc_SyntaxError, "%s\t%d\t%d\t%s",
                     PyString_AS_STRING(repr),
                     XML_GetCurrentLineNumber(parser),
                     XML_GetCurrentColumnNumber(parser),
                     XML_ErrorString(XML_GetErrorCode(parser)));
        Py_DECREF(repr);
        Py_DECREF(doc);
        Py_DECREF(stream);
      }
      return NULL;
    }
    /*See if we got pushed into the error state*/
    if (state->core.current == ERROR_STATE) {
      /*We have transitioned into the error state*/
      if (!PyErr_Occurred()) {
        /* We don't already have an exception set */
	PyObject *uri = PyObject_GetAttrString(inputSource, "uri");
	PyObject *repr = PyObject_Repr(uri);
	Py_DECREF(uri);
	PyErr_Format(g_errorObject, "%s\t%d\t%d\t%s",
		     PyString_AS_STRING(repr),
		     XML_GetCurrentLineNumber(parser),
		     XML_GetCurrentColumnNumber(parser),
		     state->core.errorMessage ? state->core.errorMessage : "Parsing aborted due to error condition"
		     );
	Py_DECREF(repr);
      }
      Py_DECREF(doc);
      Py_DECREF(stream);
      free(state->docIx);
      cleanupParserState(state);
      XML_ParserFree(parser);
      return NULL;
    }
  } while (!done);

  /*All done with the stream*/
  Py_DECREF(stream);

  doc = state->ownerDoc;
  free(state->docIx);
  if (!cleanupParserState(state)){
    /* an error occurred */
    return NULL;
  }

  XML_ParserFree(parser);

  return (PyObject *)doc;
}


void characterData(void *userData, const XML_Char *s, int len)
{
  ParserState *state = (ParserState *)userData;

  state->expat_charbuf = Unicode_FromXMLCharAndSize(s, len);

  transit(state, CHARACTER_DATA_EVENT);
  transit(state, PARSE_RESUME_EVENT);

  /*Clear up the state that we set*/
  Py_DECREF(state->expat_charbuf);
}


/*Called for the start of every new element.  Push information onto the state
  And then set the state to START_ELEMENT */
void startElement(void *userData, const XML_Char *name, const XML_Char **atts)
{
  ParserState *state = (ParserState *)userData;
  const XML_Char **ppattr;
  PyObject *curr_att_value, *curr_att_name = NULL;

  state->expat_name = Unicode_FromXMLChar(name);
  if (!state->expat_name) {
    signalError(state, "");
    return;
  }
  
  state->expat_atts = PyDict_New();
  if (!state->expat_atts) {
    Py_DECREF(state->expat_name);
    signalError(state, "");
    return;
  }

  for(ppattr = atts; *ppattr; ppattr++) {
    curr_att_name = Unicode_FromXMLChar(*ppattr++);
    curr_att_value = Unicode_FromXMLChar(*ppattr);

    PyDict_SetItem(state->expat_atts, curr_att_name, curr_att_value);

    /* let the dictionary own these */
    Py_DECREF(curr_att_value);
    Py_DECREF(curr_att_name);
  }

  if ((state->curr_elem_depth_event = (int *)calloc(1, sizeof(int))) == NULL) {
    /* Exception: Out of memory */
    return;
  }
  transit(state, START_ELEMENT_EVENT);
  transit(state, PARSE_RESUME_EVENT);

  /*Clear up the state that we set*/
  Py_DECREF(state->expat_name);
  Py_DECREF(state->expat_atts);
}


void endElement(void *userData, const XML_Char *name)
{
  ParserState *state = (ParserState *)userData;

  state->expat_name = Unicode_FromXMLChar(name);
  transit(state, END_ELEMENT_EVENT);
  transit(state, PARSE_RESUME_EVENT);

  /*Clear up the state that we set*/
  Py_DECREF(state->expat_name);
}


void startNsScope(void *userData, const XML_Char *prefix, const XML_Char *uri)
{
  ParserState *state = (ParserState *)userData;

  if (prefix) {
    state->expat_prefix = Unicode_FromXMLChar(prefix);
  } else {
    state->expat_prefix = Py_None;
    Py_INCREF(Py_None);
  }
  if (uri) {
    state->expat_uri = Unicode_FromXMLChar(uri);
  } else {
    state->expat_uri = Py_None;
    Py_INCREF(Py_None);
  }

  transit(state, START_NS_SCOPE_EVENT);
  transit(state, PARSE_RESUME_EVENT);

  /*Clear up the state that we set*/
  Py_DECREF(state->expat_prefix);
  Py_DECREF(state->expat_uri);

}


void endNsScope(void *userData, const XML_Char *prefix)
{
  ParserState *state = (ParserState *)userData;

  if (prefix) {
    state->expat_prefix = Unicode_FromXMLChar(prefix);
  } else {
    state->expat_prefix = Py_None;
    Py_INCREF(Py_None);
  }

  transit(state, END_NS_SCOPE_EVENT);
  transit(state, PARSE_RESUME_EVENT);

  /*Clear up the state that we set*/
  Py_DECREF(state->expat_prefix);
}


void processingInstruction(void *userData, const XML_Char *target, const XML_Char *data)
{
  ParserState *state = (ParserState *)userData;

  /* ignore processing instructions within the DTD */
  if (!state->in_dtd) {
    state->expat_target = Unicode_FromXMLChar(target);
    state->expat_data = Unicode_FromXMLChar(data);

    transit(state, PI_EVENT);
    transit(state, PARSE_RESUME_EVENT);

    /*Clear up the state that we set*/
    Py_DECREF(state->expat_target);
    Py_DECREF(state->expat_data);
  }
}


void comment(void *userData, const XML_Char *data)
{
  ParserState *state = (ParserState *)userData;

  /* ignore comments within the DTD */
  if (!state->in_dtd) {
    state->expat_data = Unicode_FromXMLChar(data);

    transit(state, COMMENT_EVENT);
    transit(state, PARSE_RESUME_EVENT);

    /*Clear up the state that we set*/
    Py_DECREF(state->expat_data);
  }
}


static void startDoctypeDecl(void *userData, const XML_Char *name,
                             const XML_Char *sysid, const XML_Char *pubid,
                             int has_internal_subset)
{
  ParserState *state = (ParserState *)userData;

  /* We don't do anything with the declaration, yet */
  
  state->in_dtd = 1;
}


static void endDoctypeDecl(void *userData)
{
  ParserState *state = (ParserState *)userData;
  
  state->in_dtd = 0;
}


size_t readFromObject(PyObject *object, char *buf, int max)
/* Read max bytes from a Python file-like object into buf */
{
  size_t len;
  PyObject *str = NULL;

  if (!(str = PyObject_CallMethod(object, "read", "i", max))) {
    return 0;
  }
  /*Py_INCREF(str);*/
  if (!(PyString_Check(str))) {
    /* More usable error here */
    return 0;
  }

  len = PyString_GET_SIZE(str);
  memcpy(buf, PyString_AsString(str), len);
  Py_DECREF(str);
  return len;
}


int entityRef(XML_Parser parser, const XML_Char *context,
              const XML_Char *base, const XML_Char *systemId,
              const XML_Char *publicId)
{
  char buf[BUFSIZ];
  XML_Parser new_parser;
  ParserState *state = (ParserState *)XML_GetUserData(parser);
  PyObject *stream, *new_input_source, *old_input_source, *newUri, *newUriUnicode;
  PyObject *sid_obj, *pid_obj, *hint;
  size_t len;
  int done;

  /* Invoke the entity resolver to retrieve a file-like object */
  sid_obj = Unicode_FromXMLChar(systemId);
  if (publicId) {
    pid_obj = Unicode_FromXMLChar(publicId);
  } else {
    pid_obj = Py_None;
    Py_INCREF(Py_None);
  }
  
  hint = PyString_FromString("ENTITY");

  new_input_source = PyObject_CallMethod(state->input_source, "resolve", "OOO",
					 sid_obj, pid_obj, hint);

  /* No longer needed in this function */
  Py_DECREF(sid_obj);
  Py_DECREF(pid_obj);
  Py_DECREF(hint);

  if (!new_input_source) {
    signalError(state, "");
    return 0;
  }

  stream = PyObject_GetAttrString(new_input_source, "stream");
  newUri = PyObject_GetAttrString(new_input_source, "uri");

  old_input_source = state->input_source;
  state->input_source = new_input_source;
  
  newUriUnicode = PyUnicode_FromObject(newUri);
  Py_DECREF(newUri);

  /* EXPAT uses UTF-16, Python could be UCS-4 */
  newUri = PyUnicode_EncodeUTF16(PyUnicode_AS_UNICODE(newUriUnicode),
                                 PyUnicode_GET_SIZE(newUriUnicode),
                                 NULL, (XML_BYTE_ORDER == 12) ? -1 : 1);
  Py_DECREF(newUriUnicode);

  new_parser = XML_ExternalEntityParserCreate(parser, context, NULL);
  XML_SetUserData(new_parser, state);
  
  if (!XML_SetBase(new_parser, (XML_Char *)PyString_AS_STRING(newUri)))
  {
    Py_DECREF(newUri);
    Py_DECREF(stream);
    Py_DECREF(new_input_source);
    PyErr_NoMemory();
    XML_ParserFree(new_parser);
    signalError(state, "");
    return 0;
  }
  Py_DECREF(newUri);

#ifdef DEBUG_PARSER
  fprintf(stderr, "Begin external entity parsing.\n");
#endif
  do {
    len = readFromObject(stream, buf, BUFSIZ);
#ifdef DEBUG_PARSER
    fprintf(stderr, "ER-Buf: %s.\n", buf);
#endif
    if (PyErr_Occurred()) {
      Py_DECREF(stream);
      Py_DECREF(new_input_source);
      XML_ParserFree(new_parser);
      signalError(state, "");
      return 0;
    }
    done = len < sizeof(buf);
    if (!XML_Parse(new_parser, buf, len, done)) {
      /* Exception: Well-formedness error */
      /* PyErr_Format(g_readerErrorObject, ...);*/

#ifdef DEBUG_PARSER
      fprintf(stderr, "Entity parsing failed.\n");
#endif

      if (!PyErr_Occurred()) {
	PyErr_Format(PyExc_SyntaxError, "%d:%d:%s",
		     XML_GetCurrentLineNumber(new_parser),
		     XML_GetCurrentColumnNumber(new_parser),
		     XML_ErrorString(XML_GetErrorCode(new_parser)));
      }
      Py_DECREF(stream);
      Py_DECREF(new_input_source);
      XML_ParserFree(new_parser);
      signalError(state, "");
      return 0;
    }
  } while (!done);
#ifdef DEBUG_PARSER
  fprintf(stderr, "End external entity parsing.\n");
#endif
  state->input_source = old_input_source;
  Py_DECREF(stream);
  Py_DECREF(new_input_source);
  XML_ParserFree(new_parser);

  return 1;
}


